/*
 * Gnome Memory Blocks
 * (C) 2000 the Free Software Foundation
 *
 * Author: Vova Babin <vb@dwuj.ichf.edu.pl>
 */

#include "game-app.h"

#define GAME_CLOCK_NAME "clock"
#define GAME_LOGO_FILE_NAME "gnome-memblocks-logo.png"

static GnomeAppClass *parent_class = NULL;

enum
{
  NEW_GAME,
  PREFS,
  EXIT_GAME,
  DEMO_ACTIVATED,
  SOUND_ACTIVATED,
  LAST_SIGNAL
};

static guint game_app_signals[LAST_SIGNAL] = { 0 };

static void
game_app_finalize (GtkObject * obj)
{
  GameApp *app;

  g_return_if_fail (obj != NULL);
  g_return_if_fail (IS_GAME_APP (obj));

  app = (GameApp *) obj;
  gnome_config_push_prefix (GAME_CONFIG_PREFIX);
  gnome_config_set_bool ("sound", app->sound);
  gnome_config_set_bool ("demo", app->demo);
  gnome_config_pop_prefix ();

  GTK_OBJECT_CLASS (parent_class)->finalize (obj);
}

static void
game_app_class_init (GameAppClass * klass)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass *) klass;
  object_class->finalize = game_app_finalize;

  parent_class = gtk_type_class (gnome_app_get_type ());

  klass->new_game = NULL;
  klass->prefs = NULL;
  klass->exit_game = NULL;
  klass->demo_activated = NULL;
  klass->sound_activated = NULL;

  game_app_signals[NEW_GAME] = gtk_signal_new ("new_game",
					       GTK_RUN_FIRST,
					       object_class->type,
					       GTK_SIGNAL_OFFSET
					       (GameAppClass, new_game),
					       gtk_marshal_NONE__NONE,
					       GTK_TYPE_NONE, 0);

  game_app_signals[PREFS] = gtk_signal_new ("prefs",
					    GTK_RUN_FIRST,
					    object_class->type,
					    GTK_SIGNAL_OFFSET (GameAppClass,
							       prefs),
					    gtk_marshal_NONE__NONE,
					    GTK_TYPE_NONE, 0);

  game_app_signals[EXIT_GAME] = gtk_signal_new ("exit_game",
						GTK_RUN_FIRST,
						object_class->type,
						GTK_SIGNAL_OFFSET
						(GameAppClass, exit_game),
						gtk_marshal_NONE__NONE,
						GTK_TYPE_NONE, 0);

  game_app_signals[DEMO_ACTIVATED] = gtk_signal_new ("demo_activated",
						     GTK_RUN_FIRST,
						     object_class->type,
						     GTK_SIGNAL_OFFSET
						     (GameAppClass,
						      demo_activated),
						     gtk_marshal_NONE__NONE,
						     GTK_TYPE_NONE, 0);

  game_app_signals[SOUND_ACTIVATED] = gtk_signal_new ("sound_activated",
						      GTK_RUN_FIRST,
						      object_class->type,
						      GTK_SIGNAL_OFFSET
						      (GameAppClass,
						       sound_activated),
						      gtk_marshal_NONE__NONE,
						      GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (object_class, game_app_signals, LAST_SIGNAL);
}

/*
 *  GUI callbacks, menu stuff, etc
 */

static void new_cb (GtkWidget *, gpointer);
static void scores_cb (GtkWidget *, gpointer);
static void prefs_cb (GtkWidget *, gpointer);
static void about_cb (GtkWidget *, gpointer);
static void sound_cb (GtkWidget *, gpointer);
static void demo_cb (GtkWidget *, gpointer);
static void exit_cb (GtkWidget *, gpointer);

static GnomeUIInfo game_menu[] = {
  GNOMEUIINFO_MENU_NEW_GAME_ITEM (new_cb, NULL),
  GNOMEUIINFO_MENU_SCORES_ITEM (scores_cb, NULL),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_MENU_EXIT_ITEM (exit_cb, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo settings_menu[] = {
  GNOMEUIINFO_MENU_PREFERENCES_ITEM (prefs_cb, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo help_menu[] = {
  GNOMEUIINFO_HELP (GAME_APPNAME),
  GNOMEUIINFO_MENU_ABOUT_ITEM (about_cb, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo main_menu[] = {
  GNOMEUIINFO_MENU_GAME_TREE (game_menu),
  GNOMEUIINFO_MENU_SETTINGS_TREE (settings_menu),
  GNOMEUIINFO_MENU_HELP_TREE (help_menu),
  GNOMEUIINFO_END
};

static GnomeUIInfo toolbar_uiinfo[] = {
  {GNOME_APP_UI_ITEM, N_("New"), N_("Start new game"), new_cb,
   "toolbar/new", NULL,
   GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_NEW, 0, 0, NULL},

  {GNOME_APP_UI_TOGGLEITEM, N_("Sound"), N_("Toggle sound"), sound_cb,
   "toolbar/sound", NULL,
   GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_VOLUME, 0, 0, NULL},

  {GNOME_APP_UI_TOGGLEITEM, N_("Demo"), N_("Demo mode"), demo_cb,
   "toolbar/demo", NULL,
   GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_EXEC, 0, 0, NULL},

  {GNOME_APP_UI_ITEM, N_("Prefs"), N_("Prefernces"), prefs_cb,
   "toolbar/prefs", NULL,
   GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_PREFERENCES, 0, 0, NULL},
  {GNOME_APP_UI_ENDOFINFO}
};

static void
exit_dialog_cb (gint reply, gpointer data)
{
  if (reply == GNOME_OK)
    gtk_signal_emit (GTK_OBJECT (data), game_app_signals[EXIT_GAME]);
}

static gint
delete_cb (GtkWidget * w, GdkEventAny * e, gpointer data)
{
  gnome_question_dialog_modal_parented (N_("Really exit ?"),
					exit_dialog_cb,
					data, GTK_WINDOW (data));
  return TRUE;
}

static void
prefs_cb (GtkWidget * w, gpointer data)
{
  gtk_signal_emit (GTK_OBJECT (data), game_app_signals[PREFS]);
}

static void
new_cb_cb (gint reply, gpointer data)
{
  if (reply == GNOME_OK && !GAME_APP (data)->demo)
    gtk_signal_emit (GTK_OBJECT (data), game_app_signals[NEW_GAME]);
}

static void
new_cb (GtkWidget * w, gpointer data)
{
  gnome_question_dialog_modal_parented (N_("Start new game ?"),
					new_cb_cb, data, GTK_WINDOW (data));
}

static void
exit_cb (GtkWidget * w, gpointer data)
{
  gnome_question_dialog_modal_parented (N_("Really exit ?"),
					exit_dialog_cb,
					data, GTK_WINDOW (data));
}

static void
scores_cb (GtkWidget * w, gpointer data)
{
  gnome_scores_display (_(GAME_APPNAME_LONG), GAME_APPNAME, "default", 0);
}

static void
sound_cb (GtkWidget * w, gpointer data)
{
  GameApp *app = GAME_APP (data);

  g_return_if_fail (app != NULL);

  app->sound = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (w));
  gtk_signal_emit ((GtkObject *) app, game_app_signals[SOUND_ACTIVATED]);
}

static void
demo_cb (GtkWidget * w, gpointer data)
{
  GameApp *app = GAME_APP (data);

  g_return_if_fail (app != NULL);

  app->demo = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (w));
  gtk_widget_set_sensitive (app->new_game_menu_item, !app->demo);
  gtk_widget_set_sensitive (app->new_game_button, !app->demo);
  gtk_signal_emit ((GtkObject *) app, game_app_signals[DEMO_ACTIVATED]);
}

static void
about_cb (GtkWidget * w, gpointer data)
{
  GtkWidget *about;
  const gchar *authors[] = {
    "Code: Vova Babin <vb@dwuj.ichf.edu.pl>,",
    "      Svyatoslav Kondrat <valiska@lulu.ichf.edu.pl> (clock)",
    "Sounds: Pawel Ciach <pciach@sgh.waw.pl>",
    NULL
  };

  about = gnome_about_new (_("Gnome Memory Blocks"),
			   VERSION,
			   "(C) 2000 The Free Software Foundation",
			   (const gchar **) authors,
			   _("Send comments and bug reports to:\n"
			     "             vb@dwuj.ichf.edu.pl"),
			   GAME_LOGO_FILE_NAME);

  gnome_dialog_set_parent (GNOME_DIALOG (about), GTK_WINDOW (data));
  gtk_widget_show (about);
}

static void
game_app_clock_layout_changed (GnomeDock * dock, GameClock * clock)
{
  GnomeDockItem *dock_item;
  GnomeDockPlacement place;
  GameDigitDirection direction;

  dock_item = gnome_dock_get_item_by_name (dock,
					   GAME_CLOCK_NAME,
					   &place, NULL, NULL, NULL);
  g_assert (dock_item);

  if (place != GNOME_DOCK_RIGHT)
    direction = GAME_DIGIT_DIRECTION_NORMAL;
  else
    direction = GAME_DIGIT_DIRECTION_INVERSE;

  game_clock_set_direction (clock, direction);
}

static void
game_app_init (GameApp * game)
{
  GnomeApp *app;
  GnomeDockItem *gdi;

  g_return_if_fail (game != NULL);
  g_return_if_fail (IS_GAME_APP (game));

  app = GNOME_APP (game);
  gnome_app_construct (app, GAME_APPNAME, GAME_APPNAME_LONG);
  gnome_app_create_menus_with_data (app, main_menu, game);
  game->new_game_menu_item = game_menu[0].widget;

  gnome_app_create_toolbar_with_data (app, toolbar_uiinfo, game);
  game->new_game_button = toolbar_uiinfo[0].widget;
  game->demo_toggle_button = toolbar_uiinfo[2].widget;
  game->sound_toggle_button = toolbar_uiinfo[1].widget;
  gdi = gnome_app_get_dock_item_by_name (app, GNOME_APP_TOOLBAR_NAME);
  gdi->behavior = GNOME_DOCK_ITEM_BEH_NORMAL;

  gnome_config_push_prefix (GAME_CONFIG_PREFIX);
  game_app_set_sound (game, gnome_config_get_bool ("sound=true"));
  game_app_set_demo (game, gnome_config_get_bool ("demo=true"));
  gnome_config_pop_prefix ();

  game->bar =
    (GnomeAppBar *) gnome_appbar_new (FALSE, TRUE, GNOME_PREFERENCES_USER);
  gnome_app_set_statusbar (app, GTK_WIDGET (game->bar));
  gnome_app_install_appbar_menu_hints (game->bar, main_menu);

  game->board = game_board_new ();

  gnome_app_set_contents (app, game->board);
  gtk_signal_connect (GTK_OBJECT (app), "delete_event",
		      GTK_SIGNAL_FUNC (delete_cb), game);

  game->clock = GAME_CLOCK (game_clock_new (GAME_CLOCK_INCREASING));
  gnome_app_add_docked (app, GTK_WIDGET (game->clock), GAME_CLOCK_NAME, 0, 0, 1, 1, 0);
  gtk_signal_connect (GTK_OBJECT (gnome_app_get_dock (app)), "layout_changed",
		      GTK_SIGNAL_FUNC (game_app_clock_layout_changed),
		      game->clock);
  gtk_widget_show (GTK_WIDGET (game->clock));
  gtk_widget_show (game->board);
  return;
}

GtkType
game_app_get_type (void)
{
  static GtkType type = 0;

  if (!type)
    {
      static const GtkTypeInfo type_info = {
	"GameApp",
	sizeof (GameApp),
	sizeof (GameAppClass),
	(GtkClassInitFunc) game_app_class_init,
	(GtkObjectInitFunc) game_app_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      type = gtk_type_unique (gnome_app_get_type (), &type_info);
    }
  return type;
}

/*
 *   Public interface
 */

void
game_app_set_sound (GameApp * app, gboolean sound)
{
  g_return_if_fail (app != NULL);
  g_return_if_fail (IS_GAME_APP (app));

  app->sound = sound;
  gtk_signal_handler_block_by_func (GTK_OBJECT (app->sound_toggle_button),
				    sound_cb, app);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (app->sound_toggle_button),
				sound);
  gtk_signal_handler_unblock_by_func (GTK_OBJECT (app->sound_toggle_button),
				      sound_cb, app);
}

void
game_app_set_demo (GameApp * app, gboolean demo)
{
  g_return_if_fail (app != NULL);
  g_return_if_fail (IS_GAME_APP (app));

  app->demo = demo;
  gtk_signal_handler_block_by_func (GTK_OBJECT (app->demo_toggle_button),
				    demo_cb, app);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (app->demo_toggle_button),
				demo);
  gtk_widget_set_sensitive (app->new_game_menu_item, !app->demo);
  gtk_widget_set_sensitive (app->new_game_button, !app->demo);
  gtk_signal_handler_unblock_by_func (GTK_OBJECT (app->demo_toggle_button),
				      demo_cb, app);
}

GtkWidget *
game_app_new (void)
{
  return GTK_WIDGET (gtk_type_new (GAME_APP_TYPE));
}
