/* 
 * Gnome Memory Blocks
 *
 * Copyright (C) 2000 The Free Software Foundation
 *
 * Author: Vova Babin <vb@dwuj.ichf.edu.pl>
 */

#include <time.h>
#include <stdlib.h>
#include "game-board.h"

#define INIT_ITEMS_PAD 5

struct _GameBoard
{
  GtkBin bin;
  GtkWidget *table;
};

struct _GameBoardClass
{
  GtkBinClass klass;
};

/*
 *   GtkWidget interface
 */

static void
game_board_size_allocate (GtkWidget * w, GtkAllocation * a)
{
  GtkAllocation cha;

  g_return_if_fail (w != NULL);
  g_return_if_fail (a != NULL);
  g_return_if_fail (IS_GAME_BOARD (w));

  w->allocation = *a;

  if (GTK_WIDGET_REALIZED (w))
    gdk_window_move_resize (w->window, a->x, a->y, a->width, a->height);

  cha.x = GTK_CONTAINER (w)->border_width;
  cha.y = GTK_CONTAINER (w)->border_width;
  cha.width = MAX (a->width - GTK_CONTAINER (w)->border_width * 2, 0);
  cha.height = MAX (a->height - GTK_CONTAINER (w)->border_width * 2, 0);

  if (GTK_BIN (w)->child)
    gtk_widget_size_allocate (GTK_BIN (w)->child, &cha);
}

static void
game_board_size_request (GtkWidget * w, GtkRequisition * r)
{
  GtkBin *bin;
  GtkRequisition chr;

  g_return_if_fail (w != NULL);
  g_return_if_fail (r != NULL);
  g_return_if_fail (IS_GAME_BOARD (w));

  bin = GTK_BIN (w);

  r->width = GTK_CONTAINER (w)->border_width * 2;
  r->height = GTK_CONTAINER (w)->border_width * 2;

  if (bin->child && GTK_WIDGET_VISIBLE (bin->child))
    {
      gtk_widget_size_request (bin->child, &chr);
      r->width += chr.width;
      r->height += chr.height;
    }
}

static void
game_board_paint (GtkWidget * w, GdkRectangle * area)
{
  gtk_style_apply_default_background (w->style,
				      w->window,
				      TRUE,
				      GTK_STATE_NORMAL, area, 0, 0, -1, -1);
}

static void
game_board_draw (GtkWidget * w, GdkRectangle * area)
{
  GtkBin *bin;
  GdkRectangle child_area;

  g_return_if_fail (w != NULL);
  g_return_if_fail (IS_GAME_BOARD (w));

  if (GTK_WIDGET_DRAWABLE (w))
    {
      bin = GTK_BIN (w);

      game_board_paint (w, area);

      if (bin->child && gtk_widget_intersect (bin->child, area, &child_area))
	gtk_widget_draw (bin->child, &child_area);
    }
}

static gint
game_board_expose (GtkWidget * w, GdkEventExpose * e)
{
  GtkBin *bin;
  GdkEventExpose child_event;

  g_return_val_if_fail (w != NULL, FALSE);
  g_return_val_if_fail (IS_GAME_BOARD (w), FALSE);
  g_return_val_if_fail (e != NULL, FALSE);

  if (GTK_WIDGET_DRAWABLE (w))
    {
      bin = GTK_BIN (w);
      game_board_paint (w, &e->area);

      child_event = *e;
      if (bin->child &&
	  GTK_WIDGET_NO_WINDOW (bin->child) &&
	  gtk_widget_intersect (bin->child, &e->area, &child_event.area))
	gtk_widget_event (bin->child, (GdkEvent *) & child_event);
    }
  return FALSE;
}

static void
game_board_realize (GtkWidget * w)
{
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (w != NULL);
  g_return_if_fail (IS_GAME_BOARD (w));

  GTK_WIDGET_SET_FLAGS (w, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = w->allocation.x;
  attributes.y = w->allocation.y;
  attributes.width = w->allocation.width;
  attributes.height = w->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (w);
  attributes.colormap = gtk_widget_get_colormap (w);
  attributes.event_mask = gtk_widget_get_events (w);
  attributes.event_mask |= GDK_EXPOSURE_MASK;
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  w->window = gdk_window_new (gtk_widget_get_parent_window (w),
			      &attributes, attributes_mask);
  gdk_window_set_user_data (w->window, w);
  w->style = gtk_style_attach (w->style, w->window);
  gtk_style_set_background (w->style, w->window, GTK_STATE_NORMAL);
}

/*
 *  *_init ()  routines
 */

static void
game_board_class_init (GameBoardClass * klass)
{
  GtkWidgetClass *widget_class;

  g_return_if_fail (klass != NULL);
  g_return_if_fail (IS_GAME_BOARD_CLASS (klass));

  widget_class = (GtkWidgetClass *) klass;
  widget_class->realize = game_board_realize;
  widget_class->size_allocate = game_board_size_allocate;
  widget_class->size_request = game_board_size_request;
  widget_class->draw = game_board_draw;
  widget_class->expose_event = game_board_expose;
}

static void
game_board_init (GameBoard * board)
{
  GtkWidget *align;

  g_return_if_fail (board != NULL);
  g_return_if_fail (IS_GAME_BOARD (board));

  GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (board), GTK_NO_WINDOW);

  board->table = gtk_table_new (1, 1, TRUE);
  gtk_table_set_row_spacings (GTK_TABLE (board->table), 0);
  gtk_table_set_col_spacings (GTK_TABLE (board->table), 0);
  gtk_container_set_border_width (GTK_CONTAINER (board->table), 0);

  align = gtk_alignment_new (0.5, 0.5, 0.5, 0.5);
  gtk_container_add (GTK_CONTAINER (align), board->table);
  gtk_widget_show (board->table);

  gtk_container_add (GTK_CONTAINER (board), align);
  gtk_widget_show (align);
}

/*
 *  misc
 */

static void
game_board_set_item_style (GtkWidget * w, gpointer data)
{
  if (data)
    gtk_widget_set_style (w, (GtkStyle *) data);
  else
    gtk_widget_set_rc_style (w);
}

static void
game_board_foreach_callback (GtkWidget * w, gpointer data)
{
  struct CallbackData
  {
    GameBoardFunc callback;
    gpointer data;
  }
   *callback_data = (struct CallbackData *) data;

  if (IS_GAME_ITEM (w) && callback_data->callback)
    callback_data->callback (GAME_ITEM (w), callback_data->data);
}

static void
game_board_reset_and_shufle_cb (GameItem * item_A, GameBoard * board)
{
  GameItem *item_B;
  GdkPixbuf *pic_A, *pic_B;

  game_item_reset (item_A);

  item_B = game_board_get_random_item (board);

  pic_A = game_item_get_picture (item_A);
  pic_B = game_item_get_picture (item_B);
  gdk_pixbuf_ref (pic_B);
  game_item_set_picture (item_B, pic_A);
  game_item_set_picture (item_A, pic_B);
}

/*
 *  Public methods
 */

GtkType
game_board_get_type (void)
{
  static GtkType game_board_type = 0;

  if (!game_board_type)
    {
      static const GtkTypeInfo game_board_type_info = {
	"GameBoard",
	sizeof (GameBoard),
	sizeof (GameBoardClass),
	(GtkClassInitFunc) game_board_class_init,
	(GtkObjectInitFunc) game_board_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      game_board_type = gtk_type_unique (GTK_TYPE_BIN, &game_board_type_info);
    }
  return game_board_type;
}

GtkWidget *
game_board_new (void)
{
  return GTK_WIDGET (gtk_type_new (GAME_BOARD_TYPE));
}

void
game_board_set_bg_from_pixmap (GameBoard * board, GdkPixmap * bg)
{
  GtkWidget *w;
  GtkStyle *style;

  g_return_if_fail (board != NULL);
  g_return_if_fail (IS_GAME_BOARD (board));

  w = GTK_WIDGET (board);

  if (!bg)
    {
      gtk_container_foreach (GTK_CONTAINER (board->table),
			     game_board_set_item_style, NULL);
      gtk_widget_set_rc_style (w);
    }
  else
    {
      style = gtk_style_copy (gtk_widget_get_style (w));
      style->bg_pixmap[GTK_STATE_NORMAL] = bg;
      gdk_pixmap_ref (bg);
      gtk_container_foreach (GTK_CONTAINER (board->table),
			     game_board_set_item_style, style);
      gtk_widget_set_style (w, style);
    }
}

void
game_board_set_bg_from_pixbuf (GameBoard * board, GdkPixbuf * bg)
{
  GdkPixmap *bg_pixmap;

  g_return_if_fail (board != NULL);
  g_return_if_fail (IS_GAME_BOARD (board));

  if (!bg)
    {
      game_board_set_bg_from_pixmap (board, NULL);
      return;
    }

  gdk_pixbuf_render_pixmap_and_mask (bg, &bg_pixmap, NULL, 255);
  game_board_set_bg_from_pixmap (board, bg_pixmap);
  gdk_pixmap_unref (bg_pixmap);
}

void
game_board_attach_item (GameBoard * board, GameItem * item, guint x, guint y)
{
  GtkStyle *board_gtk_style;

  g_return_if_fail (board != NULL);
  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_BOARD (board));
  g_return_if_fail (IS_GAME_ITEM (item));

  gtk_table_attach (GTK_TABLE (board->table),
		    GTK_WIDGET (item),
		    x, x + 1,
		    y, y + 1,
		    GTK_EXPAND, GTK_EXPAND, INIT_ITEMS_PAD, INIT_ITEMS_PAD);
  board_gtk_style = gtk_widget_get_style (GTK_WIDGET (board));
  if (gtk_widget_get_style (GTK_WIDGET (item)) != board_gtk_style)
    gtk_widget_set_style (GTK_WIDGET (item), board_gtk_style);
}

void
game_board_foreach_item (GameBoard * board,
			 GameBoardFunc callback, gpointer data)
{
  struct CallbackData
  {
    GameBoardFunc callback;
    gpointer data;
  }
  callback_data =
  {
  0};

  g_return_if_fail (board != NULL);
  g_return_if_fail (IS_GAME_BOARD (board));

  callback_data.callback = callback;
  callback_data.data = data;

  gtk_container_foreach (GTK_CONTAINER (board->table),
			 game_board_foreach_callback, &callback_data);
}

GameItem *
game_board_get_random_item (GameBoard * board)
{
  guint n_items, i;
  GList *table_children;
  GameItem *result;
  GtkWidget *result_widget;

  g_return_val_if_fail (board != NULL, NULL);
  g_return_val_if_fail (IS_GAME_BOARD (board), NULL);

  table_children = GTK_TABLE (board->table)->children;
  if (!table_children)
    return NULL;

  n_items = g_list_length (table_children);
  i = (guint) (((gdouble) rand ()) / (RAND_MAX / n_items));
  result_widget =
    ((GtkTableChild *) g_list_nth_data (table_children, i))->widget;

  if (IS_GAME_ITEM (result_widget))
    result = (GameItem *) result_widget;
  else
    result = NULL;

  return result;
}

void
game_board_reset_and_shufle_items (GameBoard * board)
{
  g_return_if_fail (board != NULL);
  g_return_if_fail (IS_GAME_BOARD (board));

  srand (time (NULL));
  game_board_foreach_item (board,
			   (GameBoardFunc) game_board_reset_and_shufle_cb,
			   board);
}
