/*
 * Gnome Memory Blocks
 * (C) 2000 the Free Software Foundation
 *
 * Author: Svyatoslav Kondrat <valiska@lulu.ichf.edu.pl>
 */

#include "game-digit.h"

/*
 *   Game  Indicator  Widget 
 *   ----  ---------  ------ 
 */

#define XUNIT (widget->style->klass->xthickness < 3 ? 3 : widget->style->klass->xthickness)
#define YUNIT XUNIT

#define XLENGHT 3*XUNIT
#define YLENGHT 4*YUNIT

#define XWIDTH XUNIT
#define YWIDTH YUNIT

#define BARS_DRAWING_STATE GTK_STATE_PRELIGHT

struct _GameIndicator
{
  GtkWidget widget;

  GdkPixmap *offscreen_pixmap;
  GtkOrientation orientation;	/* vertical or horizontal orientation */
  GameIndicatorBars bars;	/* which bar to draw */
};

struct _GameIndicatorClass
{
  GtkWidgetClass parent_class;
};

static void game_indicator_class_init (GameIndicatorClass *);
static void game_indicator_init (GameIndicator *);
static void game_indicator_size_request (GtkWidget *, GtkRequisition *);
static void game_indicator_size_allocate (GtkWidget *, GtkAllocation *);
static void game_indicator_finalize (GtkObject *);
static void game_indicator_realize (GtkWidget *);
static gint game_indicator_expose (GtkWidget *, GdkEventExpose *);

static void create_pixmap (GameIndicator *);
static void indicator_paint (GameIndicator *);

inline static void draw_indicators_horizontal (GameIndicator *);
inline static void draw_indicators_vertical (GameIndicator *);
static void draw_indicators (GameIndicator *);

static GtkWidgetClass *parent_class = NULL;

GtkType game_indicator_get_type (void)
{
  static GtkType game_indicator_type = 0;

  if (!game_indicator_type)
    {
      static const GtkTypeInfo game_indicator_info = {
	"GameIndicator",
	sizeof (GameIndicator),
	sizeof (GameIndicatorClass),
	(GtkClassInitFunc) game_indicator_class_init,
	(GtkObjectInitFunc) game_indicator_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      game_indicator_type =
	gtk_type_unique (GTK_TYPE_WIDGET, &game_indicator_info);
    }
  return game_indicator_type;
}

static void
game_indicator_class_init (GameIndicatorClass * class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  parent_class = gtk_type_class (GTK_TYPE_WIDGET);

  object_class = (GtkObjectClass *) class;
  object_class->finalize = game_indicator_finalize;

  widget_class = (GtkWidgetClass *) class;
  widget_class->size_request = game_indicator_size_request;
  widget_class->size_allocate = game_indicator_size_allocate;
  widget_class->realize = game_indicator_realize;
  widget_class->expose_event = game_indicator_expose;
}

static void
game_indicator_init (GameIndicator * ind)
{
  g_return_if_fail (ind != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));

  GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (ind), GTK_NO_WINDOW);

  ind->offscreen_pixmap = NULL;
  ind->orientation = GTK_ORIENTATION_HORIZONTAL;
  ind->bars = GAME_INDICATOR_EMPTY;
}

static void
game_indicator_size_request (GtkWidget * widget, GtkRequisition * requisition)
{
  GameIndicator *ind;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (widget));
  g_return_if_fail (requisition != NULL);

  ind = GAME_INDICATOR (widget);

  if (ind->orientation == GTK_ORIENTATION_HORIZONTAL)
    {
      requisition->width = 2 * XUNIT + 2 * XWIDTH + XLENGHT;
      requisition->height = 2 * YUNIT + 3 * YWIDTH + 2 * YLENGHT;
    }
  else
    {
      requisition->width = 2 * XUNIT + 3 * XWIDTH + 2 * XLENGHT;
      requisition->height = 2 * YUNIT + 2 * YWIDTH + YLENGHT;
    }
}

static void
game_indicator_size_allocate (GtkWidget * widget, GtkAllocation * allocation)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

      create_pixmap (GAME_INDICATOR (widget));
    }
}

static void
game_indicator_finalize (GtkObject * object)
{
  GameIndicator *ind;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (object));

  ind = GAME_INDICATOR (object);

  if (ind->offscreen_pixmap)
    gdk_pixmap_unref (ind->offscreen_pixmap);

  GTK_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
game_indicator_realize (GtkWidget * widget)
{
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= GDK_EXPOSURE_MASK;
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
				   &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, widget);
  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);

  create_pixmap ((GameIndicator *) widget);
}

static gint
game_indicator_expose (GtkWidget * widget, GdkEventExpose * event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GAME_INDICATOR (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (GTK_WIDGET_DRAWABLE (widget))
    gdk_draw_pixmap (widget->window,
		     widget->style->black_gc,
		     ((GameIndicator *) widget)->offscreen_pixmap,
		     event->area.x, event->area.y,
		     event->area.x, event->area.y,
		     event->area.width, event->area.height);
  return FALSE;
}

static void
create_pixmap (GameIndicator * ind)
{
  GtkWidget *widget;

  g_return_if_fail (ind != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));


  if (GTK_WIDGET_REALIZED (ind))
    {
      widget = GTK_WIDGET (ind);
      if (ind->offscreen_pixmap)
	gdk_pixmap_unref (ind->offscreen_pixmap);

      ind->offscreen_pixmap = gdk_pixmap_new (widget->window,
					      widget->allocation.width,
					      widget->allocation.height, -1);

      indicator_paint (ind);
    }
}

static void
indicator_paint (GameIndicator * ind)
{
  GtkWidget *widget;

  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));

  if (!ind->offscreen_pixmap)
    return;

  widget = GTK_WIDGET (ind);
  gtk_style_apply_default_background (widget->style,
				      ind->offscreen_pixmap,
				      FALSE,
				      GTK_STATE_NORMAL, NULL, 0, 0, -1, -1);
  draw_indicators (ind);
}

static void
draw_indicators (GameIndicator * ind)
{
  g_return_if_fail (ind != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));

  switch (ind->orientation)
    {
    case GTK_ORIENTATION_VERTICAL:
      draw_indicators_vertical (ind);
      break;
    case GTK_ORIENTATION_HORIZONTAL:
      draw_indicators_horizontal (ind);
      break;
    default:
      g_assert_not_reached ();
    }
}

inline static void
draw_indicators_vertical (GameIndicator * ind)
{
  GtkWidget *widget;

  g_return_if_fail (ind != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));

  widget = GTK_WIDGET (ind);

  if ((ind->bars & GAME_INDICATOR_A) & GAME_INDICATOR_A)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT, XUNIT, YUNIT + YWIDTH, XWIDTH, YLENGHT);

  if ((ind->bars & GAME_INDICATOR_B) & GAME_INDICATOR_B)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + 2 * XWIDTH + 2 * XLENGHT,
		  YUNIT + YWIDTH, XWIDTH, YLENGHT);

  if ((ind->bars & GAME_INDICATOR_C) & GAME_INDICATOR_C)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + XWIDTH, YUNIT + YWIDTH + YLENGHT, XLENGHT, YWIDTH);

  if ((ind->bars & GAME_INDICATOR_D) & GAME_INDICATOR_D)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT, XUNIT + XWIDTH, YUNIT, XLENGHT, YWIDTH);

  if ((ind->bars & GAME_INDICATOR_E) & GAME_INDICATOR_E)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + 2 * XWIDTH + XLENGHT,
		  YUNIT + YWIDTH + YLENGHT, XLENGHT, YWIDTH);

  if ((ind->bars & GAME_INDICATOR_F) & GAME_INDICATOR_F)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + 2 * XWIDTH + XLENGHT, YUNIT, XLENGHT, YWIDTH);

  if ((ind->bars & GAME_INDICATOR_H) & GAME_INDICATOR_H)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + XLENGHT + XWIDTH, YUNIT + YWIDTH, XWIDTH, YLENGHT);

  if ((ind->bars & GAME_INDICATOR_O) & GAME_INDICATOR_O)
    {
      gtk_draw_box (widget->style,
		    ind->offscreen_pixmap,
		    BARS_DRAWING_STATE,
		    GTK_SHADOW_OUT,
		    (XUNIT + XLENGHT) / 2 + XWIDTH,
		    (YUNIT + YLENGHT) / 2 + YWIDTH, XUNIT, YUNIT);
      gtk_draw_box (widget->style,
		    ind->offscreen_pixmap,
		    BARS_DRAWING_STATE,
		    GTK_SHADOW_OUT,
		    0.5 * XUNIT + XWIDTH + 1.5 * XLENGHT + XWIDTH,
		    (YUNIT + YLENGHT) / 2 + YWIDTH, XUNIT, YUNIT);
    }
}

inline static void
draw_indicators_horizontal (GameIndicator * ind)
{
  GtkWidget *widget;

  g_return_if_fail (ind != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));

  widget = GTK_WIDGET (ind);

  if ((ind->bars & GAME_INDICATOR_A) & GAME_INDICATOR_A)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT, XUNIT + XWIDTH, YUNIT, XLENGHT, YWIDTH);

  if ((ind->bars & GAME_INDICATOR_B) & GAME_INDICATOR_B)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + XWIDTH,
		  2 * YLENGHT + YUNIT + 2 * YWIDTH, XLENGHT, YWIDTH);

  if ((ind->bars & GAME_INDICATOR_C) & GAME_INDICATOR_C)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT, XUNIT, YUNIT + YWIDTH, XWIDTH, YLENGHT);

  if ((ind->bars & GAME_INDICATOR_D) & GAME_INDICATOR_D)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + XLENGHT + XWIDTH, YUNIT + YWIDTH, XWIDTH, YLENGHT);

  if ((ind->bars & GAME_INDICATOR_E) & GAME_INDICATOR_E)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT, YUNIT + 2 * YWIDTH + YLENGHT, XWIDTH, YLENGHT);

  if ((ind->bars & GAME_INDICATOR_F) & GAME_INDICATOR_F)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + XLENGHT + XWIDTH,
		  YUNIT + 2 * YWIDTH + YLENGHT, XWIDTH, YLENGHT);

  if ((ind->bars & GAME_INDICATOR_H) & GAME_INDICATOR_H)
    gtk_draw_box (widget->style,
		  ind->offscreen_pixmap,
		  BARS_DRAWING_STATE,
		  GTK_SHADOW_OUT,
		  XUNIT + XWIDTH, YUNIT + YWIDTH + YLENGHT, XLENGHT, YWIDTH);

  if ((ind->bars & GAME_INDICATOR_O) & GAME_INDICATOR_O)
    {
      gtk_draw_box (widget->style,
		    ind->offscreen_pixmap,
		    BARS_DRAWING_STATE,
		    GTK_SHADOW_OUT,
		    (XUNIT + XLENGHT) / 2 + XWIDTH,
		    (YUNIT + YLENGHT) / 2 + YWIDTH, XUNIT, YUNIT);
      gtk_draw_box (widget->style,
		    ind->offscreen_pixmap,
		    BARS_DRAWING_STATE,
		    GTK_SHADOW_OUT,
		    (XUNIT + XLENGHT) / 2 + XWIDTH,
		    0.5 * YUNIT + YWIDTH + 1.5 * YLENGHT + YWIDTH,
		    XUNIT, YUNIT);
    }
}

/*
 *  Public methods  of Game Indicator Widget    
 */

GtkWidget *
game_indicator_new (GtkOrientation orientation)
{
  GtkWidget *ind;

  ind = gtk_type_new (game_indicator_get_type ());

  game_indicator_set_orientation ((GameIndicator *) ind, orientation);
  return GTK_WIDGET (ind);
}

GtkWidget *
game_indicator_new_with_bars (GtkOrientation orientation,
			      GameIndicatorBars bars)
{
  GtkWidget *ind;

  ind = game_indicator_new (orientation);
  game_indicator_set_bars ((GameIndicator *) ind, bars);
  return ind;
}

void
game_indicator_set_bars (GameIndicator * ind, GameIndicatorBars bars)
{
  g_return_if_fail (ind != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));

  if (ind->bars != bars)
    {
      ind->bars = bars;
      indicator_paint (ind);
      gtk_widget_queue_draw ((GtkWidget *) ind);
    }
}

void
game_indicator_set_orientation (GameIndicator * ind,
				GtkOrientation orientation)
{
  g_return_if_fail (ind != NULL);
  g_return_if_fail (GTK_IS_GAME_INDICATOR (ind));

  if (ind->orientation != orientation)
    {
      ind->orientation = orientation;
      gtk_widget_queue_resize (GTK_WIDGET (ind));
    }
}

/*
 *   Game  Digit  Widget
 *   ----  -----  ------
 */

struct _GameDigit
{
  GameIndicator widget;
  guint digit;
  GameDigitDirection direction;
  gboolean is_empty;
  gboolean is_semicolons;
};

struct _GameDigitClass
{
  GameIndicatorClass parent_class;
};

/* 
 * Segments representation of all digits - for using 
 * by Game Indicator Widget 
 */

static const GameIndicatorBars digits[2][10] = {
  
    {GAME_INDICATOR_A | GAME_INDICATOR_B | GAME_INDICATOR_C | GAME_INDICATOR_E
   | GAME_INDICATOR_F | GAME_INDICATOR_D,
     GAME_INDICATOR_D | GAME_INDICATOR_F,
     GAME_INDICATOR_A | GAME_INDICATOR_D | GAME_INDICATOR_H | GAME_INDICATOR_E
     | GAME_INDICATOR_B,
     GAME_INDICATOR_A | GAME_INDICATOR_D | GAME_INDICATOR_H | GAME_INDICATOR_F
     | GAME_INDICATOR_B,
     GAME_INDICATOR_C | GAME_INDICATOR_H | GAME_INDICATOR_D |
     GAME_INDICATOR_F,
     GAME_INDICATOR_A | GAME_INDICATOR_C | GAME_INDICATOR_H | GAME_INDICATOR_F
     | GAME_INDICATOR_B,
     GAME_INDICATOR_A | GAME_INDICATOR_C | GAME_INDICATOR_H | GAME_INDICATOR_F
     | GAME_INDICATOR_B | GAME_INDICATOR_E,
     GAME_INDICATOR_A | GAME_INDICATOR_D | GAME_INDICATOR_F,
     GAME_INDICATOR_A | GAME_INDICATOR_C | GAME_INDICATOR_D | GAME_INDICATOR_F
     | GAME_INDICATOR_E | GAME_INDICATOR_B | GAME_INDICATOR_H,
     GAME_INDICATOR_A | GAME_INDICATOR_C | GAME_INDICATOR_D | GAME_INDICATOR_F
     | GAME_INDICATOR_B | GAME_INDICATOR_H},
  
    {GAME_INDICATOR_A | GAME_INDICATOR_B | GAME_INDICATOR_C | GAME_INDICATOR_E
   | GAME_INDICATOR_F | GAME_INDICATOR_D,
     GAME_INDICATOR_E | GAME_INDICATOR_C,
     GAME_INDICATOR_B | GAME_INDICATOR_E | GAME_INDICATOR_H | GAME_INDICATOR_D
     | GAME_INDICATOR_A,
     GAME_INDICATOR_B | GAME_INDICATOR_E | GAME_INDICATOR_H | GAME_INDICATOR_C
     | GAME_INDICATOR_A,
     GAME_INDICATOR_F | GAME_INDICATOR_H | GAME_INDICATOR_E |
     GAME_INDICATOR_C,
     GAME_INDICATOR_B | GAME_INDICATOR_F | GAME_INDICATOR_H | GAME_INDICATOR_C
     | GAME_INDICATOR_A,
     GAME_INDICATOR_B | GAME_INDICATOR_F | GAME_INDICATOR_H | GAME_INDICATOR_C
     | GAME_INDICATOR_A | GAME_INDICATOR_D,
     GAME_INDICATOR_B | GAME_INDICATOR_E | GAME_INDICATOR_C,
     GAME_INDICATOR_A | GAME_INDICATOR_C | GAME_INDICATOR_D | GAME_INDICATOR_F
     | GAME_INDICATOR_E | GAME_INDICATOR_B | GAME_INDICATOR_H,
     GAME_INDICATOR_A | GAME_INDICATOR_C | GAME_INDICATOR_F | GAME_INDICATOR_E
     | GAME_INDICATOR_B | GAME_INDICATOR_H}
};

static void game_digit_init (GameDigit *);

GtkType game_digit_get_type (void)
{
  static GtkType game_digit_type = 0;

  if (!game_digit_type)
    {
      static const GtkTypeInfo game_digit_info = {
	"GameDigit",
	sizeof (GameDigit),
	sizeof (GameDigitClass),
	(GtkClassInitFunc) NULL,
	(GtkObjectInitFunc) game_digit_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      game_digit_type =
	gtk_type_unique (game_indicator_get_type (), &game_digit_info);
    }
  return game_digit_type;
}

static void
game_digit_init (GameDigit * gd)
{
  g_return_if_fail (gd != NULL);
  g_return_if_fail (GTK_IS_GAME_DIGIT (gd));

  gd->direction = GAME_DIGIT_DIRECTION_NORMAL;
  gd->digit = 0;
  gd->is_empty = TRUE;
  gd->is_semicolons = FALSE;
}

/*
 *  Public Function of Game Digit Widget 
 */

GtkWidget *
game_digit_new_default (void)
{
  GtkWidget *gd;

  gd = gtk_type_new (game_digit_get_type ());
  return GTK_WIDGET (gd);
}

GtkWidget *
game_digit_new (GtkOrientation orientation, GameDigitDirection direction)
{
  GtkWidget *gd;

  gd = gtk_type_new (game_digit_get_type ());
  if (orientation != GTK_ORIENTATION_HORIZONTAL)	/* default value */
    game_indicator_set_orientation (GAME_INDICATOR (gd), orientation);

  if (direction != GAME_DIGIT_DIRECTION_NORMAL)	/* default value */
    game_digit_set_direction (GAME_DIGIT (gd), direction);

  return GTK_WIDGET (gd);
}

GtkWidget *
game_digit_new_with_digit (GtkOrientation orientation,
			   GameDigitDirection direction, guint digit)
{
  GtkWidget *gd;

  gd = game_digit_new (orientation, direction);
  game_digit_set_digit (GAME_DIGIT (gd), digit);

  return GTK_WIDGET (gd);
}

void
game_digit_set_direction (GameDigit * gd, GameDigitDirection direction)
{
  g_return_if_fail (gd != NULL);
  g_return_if_fail (GTK_IS_GAME_DIGIT (gd));

  if (gd->direction != direction)
    {
      gd->direction = direction;

      if (!gd->is_empty && !gd->is_semicolons)
	game_indicator_set_bars (GAME_INDICATOR (gd),
				 digits[direction][gd->digit]);
    }
}

void
game_digit_set_digit (GameDigit * gd, guint digit)
{
  g_return_if_fail (gd != NULL);
  g_return_if_fail (GTK_IS_GAME_DIGIT (gd));
  g_return_if_fail (digit < 10);

  if (gd->is_empty)
    gd->is_empty = FALSE;

  if (gd->is_semicolons)
    gd->is_semicolons = FALSE;

  gd->digit = digit;
  game_indicator_set_bars (GAME_INDICATOR (gd), digits[gd->direction][digit]);
}

GameDigitDirection game_digit_get_direction (GameDigit * gd)
{
  g_return_val_if_fail (gd != NULL, 0);
  g_return_val_if_fail (GTK_IS_GAME_DIGIT (gd), 0);

  return gd->direction;
}

guint game_digit_get_digit (GameDigit * gd)
{
  g_return_val_if_fail (gd != NULL, 10);
  g_return_val_if_fail (GTK_IS_GAME_DIGIT (gd), 10);

  return gd->digit;
}

void
game_digit_set_empty (GameDigit * gd)
{
  g_return_if_fail (gd != NULL);
  g_return_if_fail (GTK_IS_GAME_DIGIT (gd));

  game_indicator_set_bars (GAME_INDICATOR (gd), GAME_INDICATOR_EMPTY);

  gd->digit = 10;
  gd->is_semicolons = FALSE;
  gd->is_empty = TRUE;
}

void
game_digit_set_semicolons (GameDigit * gd)
{
  g_return_if_fail (gd != NULL);
  g_return_if_fail (GTK_IS_GAME_DIGIT (gd));

  game_indicator_set_bars (GAME_INDICATOR (gd), GAME_INDICATOR_O);

  gd->digit = 10;
  gd->is_empty = FALSE;
  gd->is_semicolons = TRUE;
}

gboolean game_digit_is_semicolons (GameDigit * gd)
{
  g_return_val_if_fail (gd != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GAME_DIGIT (gd), FALSE);

  return gd->is_semicolons;
}

gboolean game_digit_is_empty (GameDigit * gd)
{
  g_return_val_if_fail (gd != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_GAME_DIGIT (gd), FALSE);

  return gd->is_empty;
}
