/* 
 * Gnome Memory Blocks
 *
 * Copyright (C) 2000 The Free Software Foundation
 *
 * Author: Vova Babin <vb@dwuj.ichf.edu.pl>
 */

#include "game-item-animation.h"

static GtkObjectClass *parent_class = NULL;

static void
game_item_frame_finalize (GtkObject * obj)
{
  GameItemFrame *frame;

  g_return_if_fail (obj != NULL);
  g_return_if_fail (IS_GAME_ITEM_FRAME (obj));

  frame = GAME_ITEM_FRAME (obj);
  if (frame->pixbuf)
    gdk_pixbuf_unref (frame->pixbuf);

  if (parent_class->finalize)
    parent_class->finalize (obj);
}

static void
game_item_frame_init (GameItemFrame * frame)
{
  g_return_if_fail (frame != NULL);
  g_return_if_fail (IS_GAME_ITEM_FRAME (frame));

  frame->pixbuf = NULL;
  frame->delay_time = 0;
  frame->x_offset = 0;
  frame->y_offset = 0;
}

static void
game_item_frame_class_init (GameItemFrameClass * klass)
{
  GtkObjectClass *object_class;

  g_return_if_fail (klass != NULL);
  g_return_if_fail (IS_GAME_ITEM_FRAME_CLASS (klass));

  parent_class = gtk_type_class (GTK_TYPE_OBJECT);

  object_class = (GtkObjectClass *) klass;
  object_class->finalize = game_item_frame_finalize;
}

GtkType game_item_frame_get_type (void)
{
  static GtkType game_item_frame_type = 0;
  if (!game_item_frame_type)
    {
      static const GtkTypeInfo game_item_frame_type_info = {
	"GameItemFrame",
	sizeof (GameItemFrame),
	sizeof (GameItemFrameClass),
	(GtkClassInitFunc) game_item_frame_class_init,
	(GtkObjectInitFunc) game_item_frame_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      game_item_frame_type = gtk_type_unique (GTK_TYPE_OBJECT,
					      &game_item_frame_type_info);
      gtk_type_set_chunk_alloc (game_item_frame_type, 16);
    }
  return game_item_frame_type;
}

GameItemFrame *
game_item_frame_new (GdkPixbuf * pixbuf,
		     guint x_offset, guint y_offset, guint delay_time)
{
  GameItemFrame *frame;

  g_return_val_if_fail (pixbuf != NULL, NULL);

  frame = GAME_ITEM_FRAME (gtk_type_new (GAME_ITEM_FRAME_TYPE));
  frame->pixbuf = pixbuf;
  gdk_pixbuf_ref (frame->pixbuf);
  frame->x_offset = x_offset;
  frame->y_offset = y_offset;
  frame->delay_time = delay_time;

  return frame;
}

struct _GameItemAnimation
{
  GtkObject object;
  /* Number of frames */
  guint n_frames;
  /* Array of GameItemFrame structures */
  GameItemFrame **frames;
  /* bounding box size */
  guint width, height;
};

struct _GameItemAnimationClass
{
  GtkObjectClass klass;
};

static void
game_item_animation_finalize (GtkObject * obj)
{
  guint i;
  GameItemAnimation *animation;

  g_return_if_fail (obj != NULL);
  g_return_if_fail (IS_GAME_ITEM_ANIMATION (obj));

  animation = GAME_ITEM_ANIMATION (obj);
  if (animation->n_frames)
    {
      for (i = 0; i < animation->n_frames; i++)
	gtk_object_unref (GTK_OBJECT (animation->frames[i]));
      g_free (animation->frames);
    }

  if (parent_class->finalize)
    parent_class->finalize (obj);
}

static void
game_item_animation_init (GameItemAnimation * animation)
{
  g_return_if_fail (animation != NULL);
  g_return_if_fail (IS_GAME_ITEM_ANIMATION (animation));

  animation->frames = NULL;
  animation->n_frames = 0;
  animation->width = 0;
  animation->height = 0;
}

static void
game_item_animation_class_init (GameItemAnimationClass * klass)
{
  GtkObjectClass *object_class;

  g_return_if_fail (klass != NULL);
  g_return_if_fail (IS_GAME_ITEM_ANIMATION_CLASS (klass));

  parent_class = gtk_type_class (GTK_TYPE_OBJECT);

  object_class = (GtkObjectClass *) klass;
  object_class->finalize = game_item_animation_finalize;
}

GtkType game_item_animation_get_type (void)
{
  static GtkType game_item_animation_type = 0;
  if (!game_item_animation_type)
    {
      static const GtkTypeInfo game_item_animation_type_info = {
	"GameItemAnimation",
	sizeof (GameItemAnimation),
	sizeof (GameItemAnimationClass),
	(GtkClassInitFunc) game_item_animation_class_init,
	(GtkObjectInitFunc) game_item_animation_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      game_item_animation_type = gtk_type_unique (GTK_TYPE_OBJECT,
						  &game_item_animation_type_info);
      gtk_type_set_chunk_alloc (game_item_animation_type, 16);
    }
  return game_item_animation_type;
}

GameItemAnimation *
game_item_animation_new (void)
{
  return GAME_ITEM_ANIMATION (gtk_type_new (GAME_ITEM_ANIMATION_TYPE));
}

void
game_item_animation_append_frame (GameItemAnimation * animation,
				  GameItemFrame * frame)
{
  guint frame_width, frame_height;

  g_return_if_fail (animation != NULL);
  g_return_if_fail (IS_GAME_ITEM_ANIMATION (animation));
  g_return_if_fail (frame != NULL);
  g_return_if_fail (IS_GAME_ITEM_FRAME (frame));

  animation->n_frames += 1;
  animation->frames =
    g_renew (GameItemFrame *, animation->frames, animation->n_frames);
  animation->frames[animation->n_frames - 1] = frame;

  gtk_object_ref (GTK_OBJECT (frame));
  gtk_object_sink (GTK_OBJECT (frame));

  frame_width = gdk_pixbuf_get_width (frame->pixbuf);
  frame_height = gdk_pixbuf_get_height (frame->pixbuf);

  animation->width = MAX (animation->width, frame_width + frame->x_offset);
  animation->height = MAX (animation->height, frame_height + frame->y_offset);
}

guint game_item_animation_get_width (GameItemAnimation * animation)
{
  g_return_val_if_fail (animation != NULL, 0);
  g_return_val_if_fail (IS_GAME_ITEM_ANIMATION (animation), 0);

  return animation->width;
}

guint game_item_animation_get_height (GameItemAnimation * animation)
{
  g_return_val_if_fail (animation != NULL, 0);
  g_return_val_if_fail (IS_GAME_ITEM_ANIMATION (animation), 0);

  return animation->height;
}

guint game_item_animation_get_n_frames (GameItemAnimation * animation)
{
  g_return_val_if_fail (animation != NULL, 0);
  g_return_val_if_fail (IS_GAME_ITEM_ANIMATION (animation), 0);

  return animation->n_frames;
}

GameItemFrame *
game_item_animation_get_nth_frame (GameItemAnimation * animation, guint n)
{
  g_return_val_if_fail (animation != NULL, NULL);
  g_return_val_if_fail (IS_GAME_ITEM_ANIMATION (animation), NULL);
  g_return_val_if_fail (n < animation->n_frames, NULL);

  return animation->frames[n];
}
