/* 
 * Gnome Memory Blocks
 *
 * Copyright (C) 2000 The Free Software Foundation
 *
 * Author: Vova Babin <vb@dwuj.ichf.edu.pl>
 */

#include "game-item-style.h"

static GtkObjectClass *parent_class = NULL;

static GameItemAnimation *
game_item_style_real_create_flip_animation (GameItemStyle * style,
					    GdkPixbuf * picture)
{
  GameItemAnimation *animation = NULL;
  GameItemStyleClass *klass;

  g_return_val_if_fail (style != NULL, NULL);
  g_return_val_if_fail (IS_GAME_ITEM_STYLE (style), NULL);

  klass = GAME_ITEM_STYLE_CLASS (GTK_OBJECT (style)->klass);

  if (klass->create_flip_animation)
    animation = klass->create_flip_animation (style, picture);
  else
    g_error
      ("GameItemStyle: create_flip_animation() is not overriden by child");

  g_assert (animation != NULL);
  g_assert (IS_GAME_ITEM_ANIMATION (animation));
  g_assert (game_item_animation_get_n_frames (animation) > 1);

  return animation;
}

static GameItemAnimation *
game_item_style_real_create_hide_animation (GameItemStyle * style,
					    GdkPixbuf * picture)
{
  GameItemAnimation *animation;
  GameItemStyleClass *klass;

  g_return_val_if_fail (style != NULL, NULL);
  g_return_val_if_fail (IS_GAME_ITEM_STYLE (style), NULL);

  klass = GAME_ITEM_STYLE_CLASS (GTK_OBJECT (style)->klass);

  if (klass->create_hide_animation)
    animation = klass->create_hide_animation (style, picture);
  else
    animation = NULL;

  return animation;
}

static void
game_item_style_cleanup_hashes (GdkPixbuf * picture, gpointer data)
{
  GameItemStyle *style = NULL;
  GameItemAnimation *animation = NULL;

  g_assert (picture != NULL);
  g_assert (data != NULL);
  g_assert (IS_GAME_ITEM_STYLE (data));

  style = GAME_ITEM_STYLE (data);

  animation =
    (GameItemAnimation *) g_hash_table_lookup (style->flip_hash, picture);
  if (animation)
    {
      gtk_object_unref (GTK_OBJECT (animation));
      g_hash_table_remove (style->flip_hash, picture);
    }

  animation =
    (GameItemAnimation *) g_hash_table_lookup (style->hide_hash, picture);
  if (animation)
    {
      gtk_object_unref (GTK_OBJECT (animation));
      g_hash_table_remove (style->hide_hash, picture);
    }

  gdk_pixbuf_finalize (picture);
}

static void
game_item_style_hash_finalize_cb (gpointer key, gpointer value, gpointer data)
{
  gtk_object_unref (GTK_OBJECT (value));
  gdk_pixbuf_set_last_unref_handler ((GdkPixbuf *) key, NULL, NULL);
}

static void
game_item_style_finalize (GtkObject * obj)
{
  GameItemStyle *style = NULL;

  g_return_if_fail (obj != NULL);
  g_return_if_fail (IS_GAME_ITEM_STYLE (obj));

  style = GAME_ITEM_STYLE (obj);
  if (style->flip_hash)
    {
      g_hash_table_foreach (style->flip_hash,
			    game_item_style_hash_finalize_cb, NULL);
      g_hash_table_destroy (style->flip_hash);
    }
  if (style->hide_hash)
    {
      g_hash_table_foreach (style->hide_hash,
			    game_item_style_hash_finalize_cb, NULL);
      g_hash_table_destroy (style->hide_hash);
    }
  parent_class->finalize (obj);
}

static void
game_item_style_init (GameItemStyle * style)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (IS_GAME_ITEM_STYLE (style));

  style->flip_hash = g_hash_table_new (g_direct_hash, g_direct_equal);
  style->hide_hash = g_hash_table_new (g_direct_hash, g_direct_equal);
}

static void
game_item_style_class_init (GameItemStyleClass * klass)
{
  GtkObjectClass *object_class;

  g_return_if_fail (klass != NULL);
  g_return_if_fail (IS_GAME_ITEM_STYLE_CLASS (klass));

  klass->create_flip_animation = NULL;
  klass->create_hide_animation = NULL;

  object_class = (GtkObjectClass *) klass;
  object_class->finalize = game_item_style_finalize;

  parent_class = gtk_type_class (GTK_TYPE_OBJECT);
}

GtkType game_item_style_get_type (void)
{
  static GtkType game_item_style_type = 0;
  if (!game_item_style_type)
    {
      static const GtkTypeInfo game_item_style_type_info = {
	"GameItemStyle",
	sizeof (GameItemStyle),
	sizeof (GameItemStyleClass),
	(GtkClassInitFunc) game_item_style_class_init,
	(GtkObjectInitFunc) game_item_style_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      game_item_style_type = gtk_type_unique (GTK_TYPE_OBJECT,
					      &game_item_style_type_info);
    }
  return game_item_style_type;
}

GameItemAnimation *
game_item_style_get_flip_animation (GameItemStyle * style,
				    GdkPixbuf * picture)
{
  GameItemAnimation *animation;

  g_return_val_if_fail (style != NULL, NULL);
  g_return_val_if_fail (IS_GAME_ITEM_STYLE (style), NULL);
  g_return_val_if_fail (picture != NULL, NULL);

  animation =
    (GameItemAnimation *) g_hash_table_lookup (style->flip_hash, picture);
  if (animation)
    return GAME_ITEM_ANIMATION (animation);

  animation = game_item_style_real_create_flip_animation (style, picture);
  gtk_object_ref (GTK_OBJECT (animation));
  gtk_object_sink (GTK_OBJECT (animation));

  g_hash_table_insert (style->flip_hash, picture, animation);
  gdk_pixbuf_set_last_unref_handler (picture, game_item_style_cleanup_hashes,
				     style);

  return animation;
}

GameItemAnimation *
game_item_style_get_hide_animation (GameItemStyle * style,
				    GdkPixbuf * picture)
{
  GameItemAnimation *animation;

  g_return_val_if_fail (style != NULL, NULL);
  g_return_val_if_fail (IS_GAME_ITEM_STYLE (style), NULL);
  g_return_val_if_fail (picture != NULL, NULL);

  animation =
    (GameItemAnimation *) g_hash_table_lookup (style->hide_hash, picture);
  if (animation)
    return GAME_ITEM_ANIMATION (animation);

  animation = game_item_style_real_create_hide_animation (style, picture);

  if (animation)
    {
      gtk_object_ref (GTK_OBJECT (animation));
      gtk_object_sink (GTK_OBJECT (animation));
      g_hash_table_insert (style->hide_hash, picture, animation);
      gdk_pixbuf_set_last_unref_handler (picture,
					 game_item_style_cleanup_hashes,
					 style);
    }

  return animation;
}
