/* 
 * Gnome Memory Blocks
 *
 * Copyright (C) 2000 The Free Software Foundation
 *
 * Author: Vova Babin <vb@dwuj.ichf.edu.pl>
 */

#include <math.h>
#include <gtk/gtk.h>
#include "game-item.h"

const gchar *game_item_state_string[LAST_STATE] = {
  "PASSIVE",
  "ACTIVE",
  "HIDDEN",
  "IN_FLIP_TO_ACTIVE",
  "IN_FLIP_TO_PASSIVE",
  "IN_HIDE"
};

enum
{
  CLICKED,
  OPERATION_FINISHED,
  LAST_SIGNAL
};

static GtkWidgetClass *parent_class = 0;
static guint game_item_signals[LAST_SIGNAL] = { 0 };

struct _GameItem
{
  GtkWidget button;
  GameItemState state;
  GdkPixbuf *picture;
  /* mouse handling */
  guint mouse_in:1;
  guint button_down:1;
  /* rendering related issues */
  GdkPixmap *buffer;
  GameItemStyle *style;
  guint anim_pos;
  guint anim_timeout;
};

/*
 *  GtkObject interface
 */

static void
game_item_finalize (GtkObject * obj)
{
  GameItem *item = NULL;

  g_return_if_fail (obj != NULL);
  g_return_if_fail (IS_GAME_ITEM (obj));

  item = GAME_ITEM (obj);
  if (item->buffer)
    gdk_pixmap_unref (item->buffer);
  if (item->style)
    gtk_object_unref (GTK_OBJECT (item->style));
  if (item->picture)
    gdk_pixbuf_unref (item->picture);
  if (item->anim_timeout)
    gtk_timeout_remove (item->anim_timeout);

  GTK_OBJECT_CLASS (parent_class)->finalize (obj);
}

/*
 *  GtkWidget interface
 */

static void
game_item_size_request (GtkWidget * w, GtkRequisition * r)
{
  GameItem *item;
  GameItemAnimation *animation;

  g_return_if_fail (w != NULL);
  g_return_if_fail (IS_GAME_ITEM (w));
  g_return_if_fail (r != NULL);

  item = GAME_ITEM (w);

  r->width = 0;
  r->height = 0;

  if (item->picture)
    {
      if (item->style)
	{
	  animation =
	    game_item_style_get_flip_animation (item->style, item->picture);
	  r->width = game_item_animation_get_width (animation);
	  r->height = game_item_animation_get_height (animation);
	}
      else
	{
	  r->width = gdk_pixbuf_get_width (item->picture);
	  r->height = gdk_pixbuf_get_height (item->picture);
	}
    }
}

static void
game_item_paint (GameItem * item)
{
  GtkWidget *w;
  GameItemAnimation *animation;
  GameItemFrame *frame;

  w = GTK_WIDGET (item);

  gdk_gc_set_ts_origin (w->style->bg_gc[GTK_STATE_NORMAL],
			-w->allocation.x, -w->allocation.y);
  gtk_style_apply_default_background (w->style,
				      item->buffer,
				      FALSE,
				      GTK_STATE_NORMAL, NULL, 0, 0, -1, -1);
  gdk_gc_set_ts_origin (w->style->bg_gc[GTK_STATE_NORMAL], 0, 0);

  if (!item->picture || item->state == GAME_ITEM_HIDDEN)
    return;

  if (item->style)
    {
      if (item->state == GAME_ITEM_IN_HIDE)
	animation =
	  game_item_style_get_hide_animation (item->style, item->picture);
      else
	animation =
	  game_item_style_get_flip_animation (item->style, item->picture);
      frame = game_item_animation_get_nth_frame (animation, item->anim_pos);
      gdk_pixbuf_render_to_drawable_alpha (frame->pixbuf,
					   item->buffer,
					   0, 0,
					   frame->x_offset,
					   frame->y_offset,
					   gdk_pixbuf_get_width (frame->
								 pixbuf),
					   gdk_pixbuf_get_height (frame->
								  pixbuf),
					   GDK_PIXBUF_ALPHA_BILEVEL, 127,
					   GDK_RGB_DITHER_NONE, 0, 0);
    }
  else
    {
      if (item->state == GAME_ITEM_PASSIVE)
	gtk_draw_box (w->style,
		      item->buffer,
		      GTK_STATE_NORMAL,
		      GTK_SHADOW_OUT,
		      0, 0,
		      gdk_pixbuf_get_width (item->picture),
		      gdk_pixbuf_get_height (item->picture));
      else
	gdk_pixbuf_render_to_drawable_alpha (item->picture,
					     item->buffer,
					     0, 0,
					     0, 0,
					     gdk_pixbuf_get_width (item->
								   picture),
					     gdk_pixbuf_get_height (item->
								    picture),
					     GDK_PIXBUF_ALPHA_BILEVEL, 127,
					     GDK_RGB_DITHER_NONE, 0, 0);
    }
}

static void
game_item_create_buffer (GameItem * item)
{
  GtkWidget *w;

  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));

  w = GTK_WIDGET (item);
  if (GTK_WIDGET_REALIZED (w))
    {
      if (item->buffer)
	gdk_pixmap_unref (item->buffer);
      item->buffer = gdk_pixmap_new (w->window,
				     w->allocation.width,
				     w->allocation.height, -1);
      game_item_paint (item);
    }
}

static void
game_item_size_allocate (GtkWidget * w, GtkAllocation * a)
{
  g_return_if_fail (w != NULL);
  g_return_if_fail (IS_GAME_ITEM (w));
  g_return_if_fail (a != NULL);

  w->allocation = *a;

  if (GTK_WIDGET_REALIZED (w))
    {
      gdk_window_move_resize (w->window, a->x, a->y, a->width, a->height);
      game_item_create_buffer (GAME_ITEM (w));
    }
}

static void
game_item_realize (GtkWidget * w)
{
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (w != NULL);
  g_return_if_fail (IS_GAME_ITEM (w));

  GTK_WIDGET_SET_FLAGS (w, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = w->allocation.x;
  attributes.y = w->allocation.y;
  attributes.width = w->allocation.width;
  attributes.height = w->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (w);
  attributes.colormap = gtk_widget_get_colormap (w);
  attributes.event_mask = gtk_widget_get_events (w);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_ENTER_NOTIFY_MASK |
			    GDK_LEAVE_NOTIFY_MASK |
			    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  w->window = gdk_window_new (gtk_widget_get_parent_window (w),
			      &attributes, attributes_mask);
  gdk_window_set_user_data (w->window, w);
  w->style = gtk_style_attach (w->style, w->window);
  gtk_style_set_background (w->style, w->window, GTK_STATE_NORMAL);
  game_item_create_buffer (GAME_ITEM (w));
}

static gint
game_item_expose (GtkWidget * w, GdkEventExpose * e)
{
  g_return_val_if_fail (w != NULL, FALSE);
  g_return_val_if_fail (IS_GAME_ITEM (w), FALSE);
  g_return_val_if_fail (e != NULL, FALSE);

  if (GTK_WIDGET_DRAWABLE (w))
    {
      gdk_draw_pixmap (w->window,
		       w->style->black_gc,
		       GAME_ITEM (w)->buffer,
		       e->area.x, e->area.y,
		       e->area.x, e->area.y, e->area.width, e->area.height);
    }
  return FALSE;
}

static gint
game_item_release (GtkWidget * w, GdkEventButton * e)
{
  g_return_val_if_fail (w != NULL, FALSE);
  g_return_val_if_fail (IS_GAME_ITEM (w), FALSE);
  g_return_val_if_fail (e != NULL, FALSE);

  if (e->type == GDK_BUTTON_RELEASE)
    if (e->button == 1)
      {
	if (GAME_ITEM (w)->mouse_in && GAME_ITEM (w)->button_down)
	  gtk_signal_emit (GTK_OBJECT (w), game_item_signals[CLICKED]);
	GAME_ITEM (w)->button_down = FALSE;
      }

  return FALSE;
}
static gint
game_item_press (GtkWidget * w, GdkEventButton * e)
{
  g_return_val_if_fail (w != NULL, FALSE);
  g_return_val_if_fail (IS_GAME_ITEM (w), FALSE);
  g_return_val_if_fail (e != NULL, FALSE);

  if (e->type == GDK_BUTTON_PRESS)
    if (e->button == 1)
      GAME_ITEM (w)->button_down = TRUE;

  return FALSE;
}

static gint
game_item_enter (GtkWidget * w, GdkEventCrossing * e)
{
  GtkWidget *event_widget;

  g_return_val_if_fail (w != NULL, FALSE);
  g_return_val_if_fail (IS_GAME_ITEM (w), FALSE);
  g_return_val_if_fail (e != NULL, FALSE);

  event_widget = gtk_get_event_widget ((GdkEvent *) e);

  if ((event_widget == w) && (e->detail != GDK_NOTIFY_INFERIOR))
    GAME_ITEM (w)->mouse_in = TRUE;

  return FALSE;
}

static gint
game_item_leave (GtkWidget * w, GdkEventCrossing * e)
{
  GtkWidget *event_widget;

  g_return_val_if_fail (w != NULL, FALSE);
  g_return_val_if_fail (IS_GAME_ITEM (w), FALSE);
  g_return_val_if_fail (e != NULL, FALSE);

  event_widget = gtk_get_event_widget ((GdkEvent *) e);

  if ((event_widget == w) && (e->detail != GDK_NOTIFY_INFERIOR))
    {
      GAME_ITEM (w)->mouse_in = FALSE;
      GAME_ITEM (w)->button_down = FALSE;
    }
  return FALSE;
}

static void
game_item_set_gtk_style (GtkWidget * w, GtkStyle * prev_style, gpointer data)
{
  g_return_if_fail (w != NULL);
  g_return_if_fail (IS_GAME_ITEM (w));

  if (GTK_WIDGET_REALIZED (w))
    game_item_paint (GAME_ITEM (w));
}

/*
 *  *_init ()  routines
 */

static void
game_item_init (GameItem * item)
{
  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));

  GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (item), GTK_NO_WINDOW);

  item->state = GAME_ITEM_PASSIVE;
  item->picture = NULL;

  item->mouse_in = FALSE;
  item->button_down = FALSE;

  item->buffer = NULL;
  item->style = NULL;
  item->anim_pos = 0;
  item->anim_timeout = 0;

  gtk_signal_connect (GTK_OBJECT (item),
		      "style_set",
		      GTK_SIGNAL_FUNC (game_item_set_gtk_style), NULL);
}

static void
game_item_class_init (GameItemClass * klass)
{
  GtkObjectClass *o_class;
  GtkWidgetClass *w_class;

  g_return_if_fail (klass != NULL);
  g_return_if_fail (IS_GAME_ITEM_CLASS (klass));

  parent_class = gtk_type_class (GTK_TYPE_WIDGET);

  o_class = (GtkObjectClass *) klass;
  o_class->finalize = game_item_finalize;

  w_class = (GtkWidgetClass *) klass;
  w_class->size_request = game_item_size_request;
  w_class->size_allocate = game_item_size_allocate;
  w_class->realize = game_item_realize;
  w_class->expose_event = game_item_expose;
  w_class->enter_notify_event = game_item_enter;
  w_class->leave_notify_event = game_item_leave;
  w_class->button_press_event = game_item_press;
  w_class->button_release_event = game_item_release;

  klass->clicked = NULL;
  klass->operation_finished = NULL;

  game_item_signals[CLICKED] =
    gtk_signal_new ("clicked",
		    GTK_RUN_FIRST,
		    o_class->type,
		    GTK_SIGNAL_OFFSET (GameItemClass, clicked),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  game_item_signals[OPERATION_FINISHED] =
    gtk_signal_new ("operation_finished",
		    GTK_RUN_FIRST,
		    o_class->type,
		    GTK_SIGNAL_OFFSET (GameItemClass, operation_finished),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (o_class, game_item_signals, LAST_SIGNAL);
}

static gint
game_item_animator (gpointer data)
{
  GameItem *item = NULL;
  GameItemAnimation *animation;
  GameItemFrame *frame;
  guint n_frames;

  g_return_val_if_fail (data, FALSE);

  item = GAME_ITEM (data);
  g_return_val_if_fail (item, FALSE);

  switch (item->state)
    {
    case GAME_ITEM_IN_FLIP_TO_ACTIVE:
    case GAME_ITEM_IN_FLIP_TO_PASSIVE:
      animation =
	game_item_style_get_flip_animation (item->style, item->picture);
      n_frames =
	game_item_animation_get_n_frames (GAME_ITEM_ANIMATION (animation));
      if (item->state == GAME_ITEM_IN_FLIP_TO_PASSIVE)
	{
	  if (item->anim_pos == 0)
	    item->state = GAME_ITEM_PASSIVE;
	  else
	    item->anim_pos -= 1;
	}
      else
	{
	  if (item->anim_pos == n_frames - 1)
	    item->state = GAME_ITEM_ACTIVE;
	  else
	    item->anim_pos += 1;
	}
      break;
    case GAME_ITEM_IN_HIDE:
      animation =
	game_item_style_get_hide_animation (item->style, item->picture);
      n_frames =
	game_item_animation_get_n_frames (GAME_ITEM_ANIMATION (animation));
      if (item->anim_pos == n_frames - 1)
	item->state = GAME_ITEM_HIDDEN;
      else
	item->anim_pos += 1;
      break;
    default:
      animation = NULL;
      g_assert_not_reached ();
    }

  if (item->state == GAME_ITEM_ACTIVE ||
      item->state == GAME_ITEM_PASSIVE || item->state == GAME_ITEM_HIDDEN)
    {
      item->anim_timeout = 0;
      gtk_signal_emit (GTK_OBJECT (item),
		       game_item_signals[OPERATION_FINISHED]);
    }
  else
    {
      frame = game_item_animation_get_nth_frame (animation, item->anim_pos);
      item->anim_timeout =
	gtk_timeout_add (frame->delay_time, game_item_animator, item);
    }

  game_item_paint (item);
  gtk_widget_draw (GTK_WIDGET (item), NULL);

  return FALSE;
}

static void
game_item_start_animation (GameItem * item)
{
  GameItemFrame *frame;
  GameItemAnimation *animation;
  guint n_frames;

  switch (item->state)
    {
    case GAME_ITEM_IN_FLIP_TO_ACTIVE:
    case GAME_ITEM_IN_FLIP_TO_PASSIVE:
      animation =
	game_item_style_get_flip_animation (item->style, item->picture);
      n_frames =
	game_item_animation_get_n_frames (GAME_ITEM_ANIMATION (animation));
      if (item->state == GAME_ITEM_IN_FLIP_TO_PASSIVE)
	item->anim_pos = n_frames - 2;
      else
	item->anim_pos = 1;
      break;
    case GAME_ITEM_IN_HIDE:
      animation =
	game_item_style_get_hide_animation (item->style, item->picture);
      n_frames =
	game_item_animation_get_n_frames (GAME_ITEM_ANIMATION (animation));
      item->anim_pos = 0;
      break;
    default:
      animation = NULL;
      g_assert_not_reached ();
    }

  frame = game_item_animation_get_nth_frame (animation, item->anim_pos);
  item->anim_timeout =
    gtk_timeout_add (frame->delay_time, game_item_animator, item);
  game_item_paint (item);
  gtk_widget_draw (GTK_WIDGET (item), NULL);
}

/*
 *  Public methods
 */

GtkType game_item_get_type (void)
{
  static GtkType game_item_type = 0;
  if (!game_item_type)
    {
      static const GtkTypeInfo game_item_type_info = {
	"GameItem",
	sizeof (GameItem),
	sizeof (GameItemClass),
	(GtkClassInitFunc) game_item_class_init,
	(GtkObjectInitFunc) game_item_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      game_item_type = gtk_type_unique (GTK_TYPE_WIDGET,
					&game_item_type_info);
      gtk_type_set_chunk_alloc (game_item_type, 16);
    }
  return game_item_type;
}

GameItem *
game_item_new (void)
{
  return GAME_ITEM (gtk_type_new (GAME_ITEM_TYPE));
}

void
game_item_set_style (GameItem * item, GameItemStyle * style)
{
  GameItemAnimation *animation;
  GameItemAnimation *prev_animation;
  GameItemStyle *prev_style;
  guint n_frames = 0, prev_n_frames = 0;
  gdouble fn_frames, fprev_n_frames, fanim_pos;

  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));
  g_return_if_fail (style != NULL);
  g_return_if_fail (IS_GAME_ITEM_STYLE (style));

  prev_style = item->style;
  item->style = style;

  if (!prev_style)
    {
      gtk_object_ref (GTK_OBJECT (item->style));
      gtk_object_sink (GTK_OBJECT (item->style));
      gtk_widget_queue_resize (GTK_WIDGET (item));
      return;
    }

  switch (item->state)
    {
    case GAME_ITEM_ACTIVE:
    case GAME_ITEM_PASSIVE:
    case GAME_ITEM_IN_FLIP_TO_ACTIVE:
    case GAME_ITEM_IN_FLIP_TO_PASSIVE:
      animation =
	game_item_style_get_flip_animation (item->style, item->picture);
      prev_animation =
	game_item_style_get_flip_animation (prev_style, item->picture);
      break;
    case GAME_ITEM_HIDDEN:
    case GAME_ITEM_IN_HIDE:
      animation =
	game_item_style_get_hide_animation (item->style, item->picture);
      prev_animation =
	game_item_style_get_hide_animation (prev_style, item->picture);
      break;
    default:
      animation = NULL;
      prev_animation = NULL;
      g_assert_not_reached ();
    }
  if (animation)
    n_frames =
      game_item_animation_get_n_frames (GAME_ITEM_ANIMATION (animation));
  if (prev_animation)
    prev_n_frames =
      game_item_animation_get_n_frames (GAME_ITEM_ANIMATION (prev_animation));

  gtk_object_unref (GTK_OBJECT (prev_style));
  gtk_object_ref (GTK_OBJECT (item->style));
  gtk_object_sink (GTK_OBJECT (item->style));

  if (item->anim_pos == 0)
    item->anim_pos = 0;
  else if (item->anim_pos == prev_n_frames - 1)
    item->anim_pos = n_frames - 1;
  else
    {
      fn_frames = (gdouble) n_frames;
      fprev_n_frames = (gdouble) prev_n_frames;
      fanim_pos = (gdouble) item->anim_pos;
      item->anim_pos = (guint) rint (fn_frames * fanim_pos / fprev_n_frames);
    }

  gtk_widget_queue_resize (GTK_WIDGET (item));
}

const GameItemStyle *
game_item_get_style (const GameItem * item)
{
  g_return_val_if_fail (item != NULL, NULL);
  g_return_val_if_fail (IS_GAME_ITEM (item), NULL);

  return item->style;
}

void
game_item_set_picture (GameItem * item, GdkPixbuf * picture)
{
  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));

  if (item->picture)
    gdk_pixbuf_unref (item->picture);

  item->picture = picture;
  gdk_pixbuf_ref (picture);

  gtk_widget_queue_resize (GTK_WIDGET (item));
}

GdkPixbuf *
game_item_get_picture (GameItem * item)
{
  g_return_val_if_fail (item != NULL, NULL);
  g_return_val_if_fail (IS_GAME_ITEM (item), NULL);

  return item->picture;
}

GameItemState game_item_get_state (GameItem * item)
{
  g_return_val_if_fail (item != NULL, -1);
  g_return_val_if_fail (IS_GAME_ITEM (item), -1);

  return item->state;
}

void
game_item_flip (GameItem * item)
{
  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));
  g_return_if_fail (item->picture != NULL);

  if (item->state == GAME_ITEM_ACTIVE)
    {
      if (item->style)
	{
	  item->state = GAME_ITEM_IN_FLIP_TO_PASSIVE;
	  game_item_start_animation (item);
	}
      else
	{
	  item->state = GAME_ITEM_PASSIVE;
	  game_item_paint (item);
	  gtk_widget_draw (GTK_WIDGET (item), NULL);
	  gtk_signal_emit (GTK_OBJECT (item),
			   game_item_signals[OPERATION_FINISHED]);
	}
    }
  else if (item->state == GAME_ITEM_PASSIVE)
    {
      if (item->style)
	{
	  item->state = GAME_ITEM_IN_FLIP_TO_ACTIVE;
	  game_item_start_animation (item);
	}
      else
	{
	  item->state = GAME_ITEM_ACTIVE;
	  game_item_paint (item);
	  gtk_widget_draw (GTK_WIDGET (item), NULL);
	  gtk_signal_emit (GTK_OBJECT (item),
			   game_item_signals[OPERATION_FINISHED]);
	}
    }
  else if (item->state == GAME_ITEM_IN_FLIP_TO_ACTIVE)
    item->state = GAME_ITEM_IN_FLIP_TO_PASSIVE;
  else if (item->state == GAME_ITEM_IN_FLIP_TO_PASSIVE)
    item->state = GAME_ITEM_IN_FLIP_TO_ACTIVE;
  else
    g_warning ("GameItem: ignoring \"flip\" request (current state is %s)",
	       game_item_state_string[item->state]);
}

void
game_item_hide (GameItem * item)
{
  GameItemAnimation *animation;

  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));
  g_return_if_fail (item->picture != NULL);

  switch (item->state)
    {
    case GAME_ITEM_ACTIVE:
      animation =
	(item->
	 style ? game_item_style_get_hide_animation (item->style,
						     item->picture) : NULL);
      if (animation && game_item_animation_get_n_frames (animation))
	{
	  item->state = GAME_ITEM_IN_HIDE;
	  game_item_start_animation (item);
	}
      else
	{
	  item->state = GAME_ITEM_HIDDEN;
	  game_item_paint (item);
	  gtk_widget_draw (GTK_WIDGET (item), NULL);
	  gtk_signal_emit (GTK_OBJECT (item),
			   game_item_signals[OPERATION_FINISHED]);
	}
      break;
    default:
      g_warning ("GameItem: ignoring \"hide\" request (current state is %s)",
		 game_item_state_string[item->state]);
    }
}

void
game_item_reset (GameItem * item)
{
  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));

  if (item->anim_timeout)
    {
      gtk_timeout_remove (item->anim_timeout);
      item->anim_timeout = 0;
    }

  item->state = GAME_ITEM_PASSIVE;
  item->anim_pos = 0;

  if (GTK_WIDGET_REALIZED (item))
    game_item_paint (item);

  gtk_widget_queue_draw (GTK_WIDGET (item));
}
