/*
 * Gnome Memory Blocks
 * (C) 2000 the Free Software Foundation
 *
 * Author: Vova Babin <vb@dwuj.ichf.edu.pl>
 */

#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <math.h>
#include <gmodule.h>
#include "game.h"
#include "game-item-generic-style.h"

#define GAME_DEMO_TIMEOUT 500

static GameAppClass *parent_class = NULL;

/*
 *   Helpers
 */

static void
game_notify_user_about_success (Game * game)
{
  gchar *message;

  g_return_if_fail (IS_GAME (game));

  message = g_strdup_printf (N_("Total items %d         Challenged %d"),
			     game->total_items, game->hidden_items);
  gnome_appbar_set_status (GAME_APP (game)->bar, message);
  g_free (message);
}

static void
game_play_flip_sound (Game * game)
{
  g_return_if_fail (game != NULL);
  g_return_if_fail (IS_GAME (game));

  if (GAME_APP (game)->sound)
    gnome_triggers_do ("", NULL, GAME_APPNAME, "flip", NULL);
}

static void
game_apply_config (Game * game)
{
  GameItemStyle *style;
  GameBoard *board;
  GdkPixbuf *bg_pixbuf;
  gchar *engines_directory, *engine_path;

  g_return_if_fail (game != NULL);
  g_return_if_fail (IS_GAME (game));

  board = GAME_BOARD (GAME_APP (game)->board);
  if (game->pixmap_on_bg)
    {
      bg_pixbuf = gdk_pixbuf_new_from_file (game->bg_pixmap_fname);
      game_board_set_bg_from_pixbuf (board, bg_pixbuf);
      if (bg_pixbuf)
	gdk_pixbuf_unref (bg_pixbuf);
    }
  else
    game_board_set_bg_from_pixmap (board, NULL);

  engines_directory = gnome_unconditional_libdir_file (GAME_APPNAME);
  engine_path =
    g_module_build_path (engines_directory, game->animation_engine);
  g_free (engines_directory);
  style = game_item_generic_style_new (engine_path);
  g_free (engine_path);

  if (style)
    game_board_foreach_item (board, (GameBoardFunc) game_item_set_style,
			     style);
  else
    gnome_dialog_run (GNOME_DIALOG
		      (gnome_warning_dialog
		       (_("Failed to load animation engine"))));
}

static void game_start (GameApp *);
static void game_exit (GameApp *);

static void
game_user_win_cb (gint replay, gpointer data)
{
  if (replay == GNOME_OK)
    game_start (GAME_APP (data));
  else
    game_exit (GAME_APP (data));
}

static void
game_user_win (Game * game)
{
  GameApp *app;
  gint pos;
  time_t seconds;
  gfloat score;
  gchar message[80];

  app = GAME_APP (game);
  game_clock_stop (app->clock);

  /* following code stripped from gnome mahjongg */

  seconds = app->clock->stopped;
  score = (seconds / 60) * 1.0 + (seconds % 60) / 100.0;

  if ((pos = gnome_score_log (score, "default", FALSE)))
    {
      gnome_scores_display (_(GAME_APPNAME_LONG), GAME_APPNAME, "default",
			    pos);
      g_snprintf (message, 80,
		  "Fantastic !  %.2f !\nYou have reached #%d in the Top Ten.\n\nAnother game ?",
		  score, pos);
    }
  else
    g_snprintf (message, 80,
		"Great !\nYou made it in %.2f.\n\nAnother game ?", score);

  gnome_question_dialog_modal_parented (message,
					game_user_win_cb,
					app, GTK_WINDOW (app));
}

/*
 *   finalize
 */

static void
game_finalize (GtkObject * obj)
{
  Game *game;

  g_return_if_fail (obj != NULL);
  g_return_if_fail (IS_GAME (obj));

  game = GAME (obj);
  gnome_config_push_prefix (GAME_CONFIG_PREFIX);
  gnome_config_set_bool ("pixmap_on_bg", game->pixmap_on_bg);
  gnome_config_set_string ("bg_pixmap_fname", game->bg_pixmap_fname);
  gnome_config_set_string ("animation_engine", game->animation_engine);
  gnome_config_pop_prefix ();

  g_free (game->bg_pixmap_fname);
  g_free (game->animation_engine);

  GTK_OBJECT_CLASS (parent_class)->finalize (obj);
}

static gboolean
game_demo_handler (gpointer data)
{
  GameApp *game_app = GAME_APP (data);
  GameItem *item;

  item = game_board_get_random_item (GAME_BOARD (game_app->board));

  if (game_item_get_state (item) == GAME_ITEM_PASSIVE)
    gtk_signal_emit_by_name (GTK_OBJECT (item), "clicked", NULL);
  return TRUE;
}

/*
 *   GameApp interface
 */

static void
game_start (GameApp * game_app)
{
  Game *game;

  g_return_if_fail (game_app != NULL);
  g_return_if_fail (IS_GAME (game_app));

  game = GAME (game_app);
  if (game->demo_timeout)
    {
      gtk_timeout_remove (game->demo_timeout);
      game->demo_timeout = 0;
    }

  game->hidden_items = 0;
  game->open_items[0] = NULL;
  game->open_items[1] = NULL;

  game_board_reset_and_shufle_items (GAME_BOARD (game_app->board));
  game_board_foreach_item (GAME_BOARD (game_app->board),
			   (GameBoardFunc) gtk_object_set_user_data,
			   GINT_TO_POINTER (FALSE));
  game_notify_user_about_success (game);
  game_clock_stop (game_app->clock);
  game_clock_start (game_app->clock);
  if (game_app->demo)
    game->demo_timeout =
      gtk_timeout_add (GAME_DEMO_TIMEOUT, game_demo_handler, game_app);
}

static void
game_exit (GameApp * game_app)
{
  gtk_main_quit ();
}

static void
cleanup_animator_menu_userdata (GtkWidget * w, gpointer data)
{
  g_free (gtk_object_get_user_data (GTK_OBJECT (w)));
}

static void
prefs_clicked_cb (GnomeDialog * dialog, gint button_num, gpointer data)
{
  Game *game = GAME (data);
  GtkWidget *animator_menu_active_widget;
  GtkWidget *animator_menu;
  GtkWidget *bg_check_button;
  GtkWidget *bg_entry;

  if (button_num == 0)
    {
      g_free (game->animation_engine);
      animator_menu =
	(GtkWidget *) gtk_object_get_data (GTK_OBJECT (dialog),
					   "animator_menu");
      animator_menu_active_widget =
	gtk_menu_get_active (GTK_MENU (animator_menu));
      game->animation_engine =
	g_strdup ((gchar *)
		  gtk_object_get_user_data (GTK_OBJECT
					    (animator_menu_active_widget)));
      gtk_container_foreach (GTK_CONTAINER (animator_menu),
			     cleanup_animator_menu_userdata, NULL);

      g_free (game->bg_pixmap_fname);
      bg_entry =
	(GtkWidget *) gtk_object_get_data (GTK_OBJECT (dialog), "bg_entry");
      game->bg_pixmap_fname =
	gnome_pixmap_entry_get_filename (GNOME_PIXMAP_ENTRY (bg_entry));

      bg_check_button =
	(GtkWidget *) gtk_object_get_data (GTK_OBJECT (dialog),
					   "bg_check_button");
      game->pixmap_on_bg =
	gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (bg_check_button));
    }
  gnome_dialog_close (dialog);
  game->prefs_window = NULL;
  game_apply_config (game);
}

static gboolean
prefs_cancel_cb (GtkWidget * w, GdkEventAny * e, gpointer data)
{
  Game *game = GAME (data);
  gnome_dialog_close (GNOME_DIALOG (game->prefs_window));
  game->prefs_window = NULL;
  return FALSE;
}

static void
bg_check_button_cb (GtkWidget * w, gpointer data)
{
  gtk_widget_set_sensitive (GTK_WIDGET (data),
			    gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON
							  (w)));
}

static void
game_prefs (GameApp * game_app)
{
  Game *game;
  GnomeDialog *dialog;
  GtkWidget *box, *vbox, *label, *omenu, *t, *animator_menu, *bg_check_button,
    *bg_entry;
  gchar *dname;
  gchar *s, *engine_name;
  DIR *dir;
  struct dirent *e;
  guint current_animator;

  game = GAME (game_app);

  if (game->prefs_window)
    return;

  game->prefs_window =
    gnome_dialog_new (_("Preferences"), GNOME_STOCK_BUTTON_OK,
		      GNOME_STOCK_BUTTON_CANCEL, NULL);
  dialog = GNOME_DIALOG (game->prefs_window);
  gtk_signal_connect (GTK_OBJECT (dialog), "clicked", prefs_clicked_cb, game);
  gtk_signal_connect (GTK_OBJECT (dialog), "delete_event",
		      (GtkCallback) prefs_cancel_cb, game);

  vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
  gtk_box_pack_start_defaults (GTK_BOX (dialog->vbox), vbox);

  box = gtk_hbox_new (FALSE, GNOME_PAD);
  gtk_box_pack_start_defaults (GTK_BOX (vbox), box);

  label = gtk_label_new (_("Animation Engine"));
  gtk_box_pack_start_defaults (GTK_BOX (box), label);
  animator_menu = gtk_menu_new ();
  gtk_object_set_data (GTK_OBJECT (dialog), "animator_menu", animator_menu);

  dname = gnome_unconditional_libdir_file (GAME_APPNAME);
  dir = opendir (dname);

  if (dir)
    {
      current_animator = 0;
      while ((e = readdir (dir)) != NULL)
	{
	  s = g_strdup (e->d_name);
	  if (strstr (s, ".so"))
	    {
	      engine_name = g_strndup (s + 3, strlen (s) - 6);
	      label = gtk_menu_item_new_with_label (engine_name);
	      gtk_menu_append (GTK_MENU (animator_menu), label);
	      gtk_object_set_user_data (GTK_OBJECT (label), engine_name);
	      if (!strncmp
		  (engine_name, game->animation_engine, strlen (s) - 6))
		gtk_menu_set_active (GTK_MENU (animator_menu),
				     current_animator);
	      current_animator++;
	    }
	  g_free (s);
	}
    }

  closedir (dir);
  g_free (dname);

  omenu = gtk_option_menu_new ();
  gtk_option_menu_set_menu (GTK_OPTION_MENU (omenu), animator_menu);
  gtk_box_pack_start_defaults (GTK_BOX (box), omenu);

  label = gtk_hseparator_new ();
  gtk_box_pack_start_defaults (GTK_BOX (vbox), label);

  box = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
  gtk_box_pack_start_defaults (GTK_BOX (vbox), box);
  bg_check_button =
    gtk_check_button_new_with_label (_("Pixmap on Background"));
  gtk_object_set_data (GTK_OBJECT (dialog), "bg_check_button",
		       bg_check_button);

  gtk_box_pack_start_defaults (GTK_BOX (box), bg_check_button);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (bg_check_button),
				game->pixmap_on_bg);
  bg_entry =
    gnome_pixmap_entry_new ("background", _("Background Pixmap"), TRUE);
  gtk_object_set_data (GTK_OBJECT (dialog), "bg_entry", bg_entry);

  gtk_widget_set_sensitive (GTK_WIDGET (bg_entry), game->pixmap_on_bg);
  gtk_signal_connect (GTK_OBJECT (bg_check_button), "toggled",
		      bg_check_button_cb, bg_entry);
  t = gnome_pixmap_entry_gtk_entry (GNOME_PIXMAP_ENTRY (bg_entry));
  gtk_entry_set_text (GTK_ENTRY (t),
		      game->bg_pixmap_fname ? game->bg_pixmap_fname : "");

  box = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
  gtk_box_pack_start_defaults (GTK_BOX (vbox), box);
  gtk_box_pack_start_defaults (GTK_BOX (box), bg_entry);

  gnome_dialog_set_parent (dialog, GTK_WINDOW (game_app));
  gtk_widget_show_all (game->prefs_window);
}

/*
 *   Core game callback & company
 */

static void
game_item_operation_finished_cb (GtkWidget * w, gpointer data)
{
  g_return_if_fail (IS_GAME_ITEM (w));

  if (GPOINTER_TO_INT (gtk_object_get_user_data (GTK_OBJECT (w))))
    {
      gtk_object_set_user_data (GTK_OBJECT (w), GINT_TO_POINTER (FALSE));
      game_item_hide ((GameItem *) w);
    }
}

static void
game_item_privat_hide (GameItem * item)
{
  g_return_if_fail (item != NULL);
  g_return_if_fail (IS_GAME_ITEM (item));

  if (game_item_get_state (item) == GAME_ITEM_ACTIVE)
    game_item_hide (item);
  else
    gtk_object_set_user_data (GTK_OBJECT (item), GINT_TO_POINTER (TRUE));
}

static void
game_click_on_item_cb (GtkWidget * w, gpointer data)
{
  Game *game;
  GameItem *item;

  g_return_if_fail (IS_GAME_ITEM (w));
  g_return_if_fail (IS_GAME (data));

  item = (GameItem *) w;
  game = (Game *) data;

  game_notify_user_about_success (game);

  if (game_item_get_state (item) == GAME_ITEM_HIDDEN ||
      game_item_get_state (item) == GAME_ITEM_IN_HIDE ||
      GPOINTER_TO_INT (gtk_object_get_user_data (GTK_OBJECT (item))))
    return;

  if (item == game->open_items[0])
    {
      game_play_flip_sound (game);
      game_item_flip (game->open_items[0]);
      game->open_items[0] = NULL;
      return;
    }

  if (item == game->open_items[1])
    {
      game_play_flip_sound (game);
      game_item_flip (game->open_items[1]);
      game->open_items[1] = NULL;
      return;
    }

  if (game->open_items[0] && game->open_items[1])
    {
      game_item_flip (game->open_items[0]);
      game_item_flip (game->open_items[1]);
      game->open_items[0] = NULL;
      game->open_items[1] = NULL;
    }

  if (game->open_items[1])
    {
      game->open_items[0] = game->open_items[1];
      game->open_items[1] = NULL;
    }

  game_play_flip_sound (game);
  game_item_flip (item);

  if (game->open_items[0])
    game->open_items[1] = item;
  else
    game->open_items[0] = item;

  if (game->open_items[0] && game->open_items[1])
    if (game_item_get_picture (game->open_items[0]) ==
	game_item_get_picture (game->open_items[1]))
      {
	game_item_privat_hide (game->open_items[0]);
	game_item_privat_hide (game->open_items[1]);
	game->open_items[0] = NULL;
	game->open_items[1] = NULL;
	game->hidden_items += 2;
      }

  game_notify_user_about_success (game);

  if (game->hidden_items == game->total_items)
    {
      if (GAME_APP (game)->demo)
	game_start (GAME_APP (game));
      else
	game_user_win (game);
    }
}

/*
 *   *_init() routines
 */

static void
game_init (Game * game)
{
  GdkPixbuf *picture;
  GList *pictures;
  GameItem *item;
  gboolean default_used;
  DIR *dir;
  struct dirent *e;
  guint x, y, j, ncols;
  gchar *dname, *s, *tmp;

  g_return_if_fail (game != NULL);
  g_return_if_fail (IS_GAME (game));

  gnome_config_push_prefix (GAME_CONFIG_PREFIX);
  game->pixmap_on_bg =
    gnome_config_get_bool_with_default ("pixmap_on_bg=true", &default_used);
  if (default_used)
    game->bg_pixmap_fname =
      gnome_unconditional_pixmap_file
      ("gnome-memblocks/default/background.jpg");
  else
    game->bg_pixmap_fname = gnome_config_get_string ("bg_pixmap_fname");
  game->animation_engine =
    gnome_config_get_string ("animation_engine=scaling");
  gnome_config_pop_prefix ();

  pictures = NULL;
  dname = gnome_unconditional_pixmap_file ("gnome-memblocks/default");
  dir = opendir (dname);

  if (dir)
    {
      while ((e = readdir (dir)) != NULL)
	{
	  s = g_strdup (e->d_name);
	  if (s[0] != '.' && !strstr (s, "background"))
	    {
	      tmp = g_concat_dir_and_file (dname, s);
	      picture = gdk_pixbuf_new_from_file (tmp);
	      g_free (tmp);
	      if (picture)
		pictures = g_list_append (pictures, picture);
	    }
	  g_free (s);
	}
      closedir (dir);
    }
  g_free (dname);

  if (!pictures)
    {
      gnome_dialog_run (GNOME_DIALOG
			(gnome_error_dialog
			 (_("Error loading default tiles"))));
      exit (1);
    }

  game->total_items = 2 * g_list_length (pictures);

  ncols = (guint) sqrt (((double) game->total_items));
  for (x = 0, y = 0, j = 0; j < game->total_items; j++)
    {
      item = game_item_new ();
      gtk_object_set_user_data (GTK_OBJECT (item), GINT_TO_POINTER (FALSE));
      gtk_signal_connect (GTK_OBJECT (item),
			  "operation_finished",
			  GTK_SIGNAL_FUNC (game_item_operation_finished_cb),
			  game);
      game_item_set_picture (item, g_list_nth_data (pictures, j / 2));
      gtk_signal_connect (GTK_OBJECT (item), "clicked",
			  GTK_SIGNAL_FUNC (game_click_on_item_cb), game);
      game_board_attach_item (GAME_BOARD (GAME_APP (game)->board), item, x,
			      y);
      gtk_widget_show (GTK_WIDGET (item));
      if (x < ncols - 1)
	x++;
      else
	{
	  x = 0;
	  y++;
	}
    }

  g_list_foreach (pictures, (GFunc) gdk_pixbuf_unref, NULL);
  g_list_free (pictures);

  game->prefs_window = NULL;
  game_apply_config (game);
  game->demo_timeout = 0;
  game_start (GAME_APP (game));
}

static void
game_class_init (GameClass * klass)
{
  GtkObjectClass *object_class;
  GameAppClass *game_app_class;

  g_return_if_fail (klass != NULL);
  g_return_if_fail (IS_GAME_CLASS (klass));

  parent_class = gtk_type_class (GAME_APP_TYPE);

  object_class = (GtkObjectClass *) klass;
  object_class->finalize = game_finalize;

  game_app_class = (GameAppClass *) klass;
  game_app_class->exit_game = game_exit;
  game_app_class->new_game = game_start;
  game_app_class->prefs = game_prefs;
  game_app_class->demo_activated = game_start;
}

/*
 *   Public
 */

GtkType game_get_type (void)
{
  static GtkType game_type = 0;
  if (!game_type)
    {
      static const GtkTypeInfo game_type_info = {
	"Game",
	sizeof (Game),
	sizeof (GameClass),
	(GtkClassInitFunc) game_class_init,
	(GtkObjectInitFunc) game_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      game_type = gtk_type_unique (GAME_APP_TYPE, &game_type_info);
    }
  return game_type;
}

GtkWidget *
game_new (void)
{
  return GTK_WIDGET (gtk_type_new (GAME_TYPE));
}
